from typing import List, Dict, Callable, TYPE_CHECKING, Union
from urllib.parse import urljoin, urlencode

from ..utils import clean_data
from ..base import Base

if TYPE_CHECKING:
    from .user import AsyncUser
    from .task import AsyncTask
    from . import AsyncGeoboxClient


class AsyncBase(Base):
    BASE_ENDPOINT = ''
    
    def __init__(self, api, **kwargs):
        """
        Initialize the Base class.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient client
            uuid (str, optional): The UUID of the resource
            data (dict, optional): The data of the resource
        """
        super().__init__(api, **kwargs)


    @classmethod
    async def _get_list(cls, api: 'AsyncGeoboxClient',  endpoint: str, params: dict = {}, factory_func: Callable = None, geojson: bool = False) -> Union[List['Base'], int]:
        """Get a list of resources with optional filtering and pagination"""
        query_string = urlencode(clean_data(params)) if params else ''
        endpoint = urljoin(endpoint, f'?{query_string}')
        response = await api.get(endpoint)

        if params.get('return_count'):
            return cls._get_count(response)
        
        if not response:
            return []
            
        if geojson:
            return cls._handle_geojson_response(api, response, factory_func)
            
        return [factory_func(api, item) for item in response]
    

    @classmethod
    async def _get_list_by_ids(cls, api: 'AsyncGeoboxClient', endpoint: str, params: dict = None, factory_func: Callable = None) -> List['Base']:
        """
        Internal method to get a list of resources by their IDs.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient client
            endpoint (str): The endpoint of the resource
            params (dict): Additional parameters for filtering and pagination
            factory_func (Callable): A function to create the resource object

        Returns:
            List[Base]: The list of resource objects
        """
        params = clean_data(params)
        query_string = urlencode(params)
        endpoint = urljoin(endpoint, f'?{query_string}')
        response = await api.get(endpoint)
        return [factory_func(api, item) for item in response]


    @classmethod
    async def _get_detail(cls, api: 'AsyncGeoboxClient', endpoint: str, uuid: str, params: dict = {}, factory_func: Callable = None) -> 'Base':
        """
        Internal method to get a single resource by UUID.
        
        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient client
            uuid (str): The UUID of the resource
            params (dict): Additional parameters for filtering and pagination
            factory_func (Callable): A function to create the resource object

        Returns:
            Base: The resource object
        """
        query_strings = urlencode(clean_data(params))
        endpoint = urljoin(endpoint, f'{uuid}/?{query_strings}')
        response = await api.get(endpoint)
        return factory_func(api, response)
    

    @classmethod
    async def _create(cls, api: 'AsyncGeoboxClient', endpoint: str, data: dict, factory_func: Callable = None) -> 'Base':
        """
        Internal method to create a resource.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient client
            data (dict): The data to create the resource with
            factory_func (Callable): A function to create the resource object

        Returns:
            Base: The created resource object
        """
        data = clean_data(data)
        response = await api.post(endpoint, data)
        return factory_func(api, response)
    

    async def _update(self, endpoint: str, data: dict, clean: bool = True) -> Dict:
        """
        Update the resource.

        Args:
            data (dict): The data to update the resource with
        """
        if clean:
            data = clean_data(data)

        response = await self.api.put(endpoint, data)
        self._update_properties(response)
        return response
    

    async def _delete(self, endpoint: str) -> None:
        """
        Delete the resource.
        """
        await self.api.delete(endpoint)
        self.uuid = None
        self.endpoint = None
        

    async def _share(self, endpoint: str, users: List['AsyncUser']) -> None:
        """
        Internal method to share the resource with the given user IDs.

        Args:
            users (List[AsyncUser]): The user objects to share the resource with
        """
        data = {"user_ids": [user.user_id for user in users]}
        endpoint = urljoin(endpoint, f'share/')
        await self.api.post(endpoint, data, is_json=False)


    async def _unshare(self, endpoint: str, users: List['AsyncUser']) -> None:
        """
        Internal method to unshare the resource with the given user IDs.

        Args:
            users (List[AsyncUser]): The user objects to unshare the resource with
        """
        data = {"user_ids": [user.user_id for user in users]}
        endpoint = urljoin(endpoint, f'unshare/')
        await self.api.post(endpoint, data, is_json=False)


    async def _get_shared_users(self, endpoint: str, params: dict = None) -> List['AsyncUser']:
        """
        Internal method to get the users that the resource is shared with.

        Args:
            endpoint (str): resource endpoint
            params (dict): Additional parameters for filtering and pagination

        Returns:
            List[AsyncUser]: The users that the resource is shared with
        """
        from .user import AsyncUser

        params = clean_data(params)
        query_strings = urlencode(params)
        endpoint = urljoin(endpoint, f'shared-with-users/?{query_strings}')
        response = await self.api.get(endpoint)
        return [AsyncUser(self.api, item['id'], item) for item in response]
    

    async def _get_settings(self, endpoint: str) -> Dict:
        """
        Internal method to get the settings of the resource.

        Args:
            endpoint (str): The endpoint of the resource
        """
        endpoint = urljoin(endpoint, f'settings/?f=json')
        return await self.api.get(endpoint)
    

    async def _set_settings(self, endpoint: str, data: dict) -> None:
        """
        Internal method to set the settings of the resource.

        Args:
            endpoint (str): The endpoint of the resource
            data (dict): The data to set the settings with
        """
        endpoint = urljoin(endpoint, f'settings/')
        return await self.api.put(endpoint, data)


    async def _get_task(self, response, error_message: str) -> List['AsyncUser']:
        from .task import AsyncTask  # avoid circular dependency

        if len(response) == 1 and isinstance(response, list) and response[0].get('task_id'):
            result = [await self.api.get_task(response[0].get('task_id'))]
        elif len(response) == 2 and isinstance(response, list) and (response[0].get('task_id') and response[1].get('task_id')):
            result = [await self.api.get_task(item.get('task_id')) for item in response]
        elif len(response) == 1 and isinstance(response, dict) and response.get('task_id'):
            result = [await self.api.get_task(response.get('task_id'))]
        else:
            raise ValueError(error_message)

        return result


    async def _seed_cache(self, endpoint: str, data: dict) -> List['AsyncTask']:
        """
        Internal method to cache seed the resource.

        Args:
            endpoint (str): The endpoint of the resource
            data (dict): The data to cache seed with
        """
        if data['workers'] not in [1, 2, 4, 8, 12, 16, 20, 24]:
            raise ValueError("workers must be in [1, 2, 4, 8, 12, 16, 20, 24]")
        
        data = clean_data(data)
        endpoint = urljoin(endpoint, f'cache/seed/')
        response = await self.api.post(endpoint, data)
        return await self._get_task(response, 'Failed to seed cache')


    async def _clear_cache(self, endpoint: str) -> None:
        """
        Internal method to clear the cache of the resource.

        Args:
            endpoint (str): The endpoint of the resource
        """
        endpoint = urljoin(endpoint, f'cache/clear/')
        await self.api.post(endpoint)


    async def _cache_size(self, endpoint: str) -> int:
        """
        Internal method to get the size of the cache of the resource.

        Args:
            endpoint (str): The endpoint of the resource
        """
        endpoint = urljoin(endpoint, f'cache/size/')
        return await self.api.post(endpoint)
    

    async def _update_cache(self, endpoint: str, data: Dict = {}) -> List['AsyncTask']:
        """
        Internal method to update the cache of the resource.

        Args:
            endpoint (str): The endpoint of the resource
        """
        data = clean_data(data)
        endpoint = urljoin(endpoint, 'cache/update/')
        response = await self.api.post(endpoint, data)
        return await self._get_task(response, 'Failed to update cache')



import pytest

from geobox.attachment import Attachment
from geobox.file import File
from geobox.vectorlayer import VectorLayer, LayerType
from geobox.feature import Feature
from geobox.view import VectorLayerView
from geobox.map import Map


def test_init(api, mock_attachment_data):
    attachment = Attachment(api, mock_attachment_data['id'], mock_attachment_data)
    assert attachment.name == mock_attachment_data['name']
    assert attachment.attachment_id == mock_attachment_data['id']
    assert attachment.data == mock_attachment_data
    assert attachment.endpoint == f'{Attachment.BASE_ENDPOINT}{attachment.attachment_id}/'


def test_repr(api, mock_attachment_data):
    attachment = Attachment(api, mock_attachment_data['id'], mock_attachment_data)
    assert repr(attachment) == f"Attachment(id={attachment.attachment_id}, name={attachment.name})"


def test_file(api, mock_attachment_data):
    attachment = Attachment(api, mock_attachment_data['id'], mock_attachment_data)
    file = attachment.file
    assert type(file) == File
    assert file.data == mock_attachment_data['file']


def test_get_attachments(api, mock_attachment_data, mock_vector_data):
    api.get.return_value = [mock_attachment_data, mock_attachment_data]
    layer = VectorLayer(api, mock_vector_data['uuid'], LayerType.Polygon, mock_vector_data)
    attachments = Attachment.get_attachments(api, resource=layer)
    api.get.assert_called_once_with(f'{Attachment.BASE_ENDPOINT}?resource_type=vector&resource_uuid=297fa7ca-877a-400c-8003-d65de9e791c2&skip=0&limit=10')
    assert len(attachments) == 2
    assert type(attachments[0]) == Attachment
    assert attachments[0].data == mock_attachment_data

    # error
    with pytest.raises(TypeError):
        Attachment.get_attachments(api, resource='')


def test_create_attachment(api, mock_attachment_data, mock_vector_data, mock_feature_data, mock_view_data, mock_map_data):
    api.post.return_value = mock_attachment_data
    layer = VectorLayer(api, mock_attachment_data['key'].split(':')[1], LayerType.Polygon, mock_vector_data)
    mock_feature_data['Polygon']['id'] = 1
    feature = Feature(layer, data=mock_feature_data['Polygon'])
    file = File(api, mock_attachment_data['file']['uuid'], mock_attachment_data['file'])
    attachment = Attachment.create_attachment(api,
                                        name=mock_attachment_data['name'],
                                        loc_x=mock_attachment_data['loc_x'],
                                        loc_y=mock_attachment_data['loc_y'],
                                        resource=layer,
                                        file=file,
                                        feature=feature,
                                        display_name=mock_attachment_data['display_name'],
                                        description=mock_attachment_data['description'])
    api.post.assert_called_once_with(Attachment.BASE_ENDPOINT, {'name': 'scene_(1).png', 'loc_x': 51.190273, 'loc_y': 35.71116, 'resource_type': 'vector', 'resource_uuid': '0da64b20-90bd-4e56-88f3-903a2dc97d49', 'element_id': 1, 'file_id': 6764})
    assert type(attachment) == Attachment
    assert attachment.attachment_id == mock_attachment_data['id']
    assert attachment.data == mock_attachment_data
    # view resource
    view = VectorLayerView(api, mock_view_data['uuid'], LayerType.Polygon, mock_view_data)
    attachment = Attachment.create_attachment(api,
                                        name=mock_attachment_data['name'],
                                        loc_x=mock_attachment_data['loc_x'],
                                        loc_y=mock_attachment_data['loc_y'],
                                        resource=view,
                                        file=file,
                                        feature=feature,
                                        display_name=mock_attachment_data['display_name'],
                                        description=mock_attachment_data['description'])
    assert type(attachment) == Attachment
    # map resource
    map = Map(api, mock_map_data['uuid'], mock_map_data)
    attachment = Attachment.create_attachment(api,
                                        name=mock_attachment_data['name'],
                                        loc_x=mock_attachment_data['loc_x'],
                                        loc_y=mock_attachment_data['loc_y'],
                                        resource=map,
                                        file=file,
                                        feature=feature,
                                        display_name=mock_attachment_data['display_name'],
                                        description=mock_attachment_data['description'])
    assert type(attachment) == Attachment



def test_update_attachment(api, mock_attachment_data):
    expected_data = {'name': 'updated_name', 
                'display_name': 'updated display name', 
                'description': 'updated description', 
                'loc_x': 10, 
                'loc_y': 20}
    api.put.return_value = {**mock_attachment_data, **expected_data}
    updated_data = Attachment.update_attachment(api, attachment_id=mock_attachment_data['id'],
                                                    name=expected_data['name'],
                                                    display_name=expected_data['display_name'],
                                                    description=expected_data['description'],
                                                    loc_x=expected_data['loc_x'],
                                                    loc_y=expected_data['loc_y'])
    api.put.assert_called_once_with(f"{Attachment.BASE_ENDPOINT}{mock_attachment_data['id']}", expected_data)
    assert updated_data == {**mock_attachment_data, **expected_data}


def test_update(api, mock_attachment_data, mock_vector_data):
    api.get.return_value = [mock_attachment_data, mock_attachment_data]
    layer = VectorLayer(api, mock_attachment_data['key'].split(':')[1], LayerType.Polygon, mock_vector_data)
    attachment = Attachment.get_attachments(api, resource=layer)[0]
    expected_data = {'name': 'updated_name', 
            'display_name': 'updated display name', 
            'description': 'updated description', 
            'loc_x': 10, 
            'loc_y': 20}
    updated_data = {**mock_attachment_data, **expected_data}

    api.put.return_value = updated_data
    attachment.update(name=expected_data['name'],
                        display_name=expected_data['display_name'],
                        description=expected_data['description'],
                        loc_x=expected_data['loc_x'],
                        loc_y=expected_data['loc_y'])
    api.put.assert_called_once_with(attachment.endpoint, expected_data)
    assert attachment.data == updated_data


def test_delete(api, mock_attachment_data, mock_map_data):
    api.get.return_value = [mock_attachment_data, mock_attachment_data]
    map = Map(api, mock_attachment_data['key'].split(':')[1], mock_map_data)
    attachment = Attachment.get_attachments(api, resource=map)[0]
    endpoint = attachment.endpoint
    attachment.delete()
    api.delete.assert_called_once_with(endpoint)
    assert attachment.attachment_id is None
    assert attachment.endpoint is None


def test_thumbnail(api, mock_attachment_data, mock_view_data):
    api.get.return_value = [mock_attachment_data, mock_attachment_data]
    view = VectorLayerView(api, mock_attachment_data['key'].split(':')[1], LayerType.Polygon, mock_view_data)
    attachment = Attachment.get_attachments(api, resource=view)[0]
    thumbnail_url = attachment.thumbnail
    assert thumbnail_url == f"{api.base_url}{attachment.endpoint}thumbnail"


def test_to_async(api, async_api, mock_attachment_data):
    attachment = Attachment(api, mock_attachment_data['id'], mock_attachment_data)
    async_instance = attachment.to_async(async_api)
    assert async_instance.api == async_api  


import pytest
from unittest.mock import patch

from geobox.route import Routing, RoutingGeometryType, RoutingOverviewLevel
from geobox.user import User

def test_route(api):
    api.get.return_value = {'route': 'value'}
    route = Routing.route(api, 
                            stops='53,33;56,36', 
                            alternatives=True, 
                            steps=True,
                            geometries=RoutingGeometryType.geojson,
                            overview=RoutingOverviewLevel.Full,
                            annotations=True)
    assert route == {'route': 'value'}
    api.get.assert_called_once_with('routing/route?stops=53%2C33%3B56%2C36&alternatives=True&steps=True&geometries=geojson&overview=full&annotaions=True')
    


from typing import List, Dict, Optional, TYPE_CHECKING, Union, Any
from urllib.parse import urljoin

from geobox.enums import TableExportFormat

from .base import Base
from .task import Task
from .enums import FieldType
from .exception import NotFoundError
from .utils import clean_data

if TYPE_CHECKING:
    from . import GeoboxClient
    from .user import User
    from .aio import AsyncGeoboxClient
    from .file import File
    from .aio.table import Table as AsyncTable, AsyncTableRow, AsyncTableField



class TableRow(Base):

    def __init__(self, 
        table: 'Table',
        data: Optional[Dict] = {},
    ):
        """
        Constructs all the necessary attributes for the TableRow object.

        Args:
            table (Table): The table that the row belongs to.
            data (Dict, optional): The data of the field.
        """
        super().__init__(api=table.api, data=data)
        self.table = table
        self.endpoint = urljoin(table.endpoint, f'rows/{self.id}/') if self.data.get('id') else None


    def __repr__(self) -> str:
        """
        Return a string representation of the TableRow.

        Returns:
            str: The string representation of the TableRow.
        """
        return f"TableRow(id={self.id}, table_name={self.table.data.get('name', 'None')})"


    @classmethod
    def create_row(cls, table: 'Table', **kwargs) -> 'TableRow':
        """
        Create a new row in the table.

        Each keyword argument represents a field value for the row, where:
            - The keyword is the field name
            - The value is the field value

        Args:
            table (Table): table instance

        Keyword Args: 
            **kwargs: Arbitrary field values matching the table schema.

        Returns:
            TableRow: created table row instance

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table, TableRow
            >>> client = GeoboxClient()
            >>> table = client.get_table(uuid="12345678-1234-5678-1234-567812345678")
            or
            >>> table = Table.get_table(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> row_data = {
                'field1': 'value1'
            }
            >>> row = TableRow.create_row(table, row_data)
        """
        endpoint = urljoin(table.endpoint, 'rows/')
        return cls._create(table.api, endpoint, kwargs, factory_func=lambda api, item: TableRow(table, data=item))


    @classmethod
    def get_row(cls, 
        table: 'Table', 
        row_id: int,
        user_id: Optional[int],
    ) -> 'TableRow':
        """
        Get a row by its id

        Args:
            table (Table): the table instance
            row_id (int): the row id
            user_id (int, optional): specific user. privileges required.

        Returns:
            TanbleRow: the table row instance

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table, TableRow
            >>> client = GeoboxClient()
            >>> table = client.get_table(uuid="12345678-1234-5678-1234-567812345678")
            or
            >>> table = Table.get_table(client, uuid="12345678-1234-5678-1234-567812345678")

            >>> row = TableRow.get_row(table, row_id=1)
        """
        param = {
            'f': 'json',
            'user_id': user_id
        }
        endpoint = urljoin(table.endpoint, f'rows/')
        return cls._get_detail(table.api, endpoint, uuid=row_id, params=param, factory_func=lambda api, item: TableRow(table, data=item))


    def update(self, **kwargs) -> Dict:
        """
        Update a row

        Keyword Args:
            fields to update

        Returns:
            Dict: updated row data

        Example:
            >>> from geobox import GeoboxClient
            >>> client = GeoboxClient()
            >>> table = client.get_table(uuid="12345678-1234-5678-1234-567812345678")
            >>> row = table.get_row(row_id=1)
            >>> row.update(field1='new_value')
        """
        super()._update(self.endpoint, self.data, clean=False)
        return self.data


    def delete(self) -> None:
        """
        Delete a row

        Returns: 
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> client = GeoboxClient()
            >>> table = client.get_table(uuid="12345678-1234-5678-1234-567812345678")
            >>> row = table.get_row(row_id=1)
            >>> row.delete()
        """
        super()._delete(self.endpoint)


    def to_async(self, async_client: 'AsyncGeoboxClient') -> 'AsyncTableRow':
        """
        Switch to async version of the table row instance to have access to the async methods

        Args:
            async_client (AsyncGeoboxClient): The async version of the GeoboxClient instance for making requests.

        Returns:
            AsyncTableRow: the async instance of the TableRow.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> client = GeoboxClient()
            >>> table = client.get_table(uuid="12345678-1234-5678-1234-567812345678")
            >>> row = table.get_row(row_id=1)
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     async_row = row.to_async(async_client)
        """
        from .aio.table import AsyncTableRow

        async_table = self.table.to_async(async_client=async_client)
        return AsyncTableRow(table=async_table, data=self.data)



class TableField(Base):

    def __init__(self, 
        table: 'Table',
        data_type: 'FieldType',
        field_id: int = None,
        data: Optional[Dict] = {},
    ):
        """
        Constructs all the necessary attributes for the Field object.

        Args:
            table (Table): The table that the field belongs to.
            data_type (FieldType): type of the field
            field_id (int): the id of the field
            data (Dict, optional): The data of the field.
        """
        super().__init__(api=table.api, data=data)
        self.table = table
        self.field_id = field_id
        if not isinstance(data_type, FieldType):
            raise ValueError("data_type must be a FieldType instance")
        self.data_type = data_type
        self.endpoint = urljoin(table.endpoint, f'fields/{self.id}/') if self.data.get('id') else None


    def __repr__(self) -> str:
        """
        Return a string representation of the field.

        Returns:
            str: The string representation of the field.
        """
        return f"TableField(id={self.id}, name={self.name}, data_type={self.data_type})"

        
    def __getattr__(self, name: str) -> Any:
        """
        Get an attribute from the resource.

        Args:
            name (str): The name of the attribute
        """
        if name == 'datatype':
            return FieldType(self.data['datatype'])
        return super().__getattr__(name)


    @property
    def domain(self) -> Dict:
        """
        Domain property

        Returns:
            Dict: domain data
        """
        return self.data.get('domain')


    @domain.setter
    def domain(self, value: Dict) -> None:
        """
        Domain property setter

        Returns:
            None
        """
        self.data['domain'] = value
        

    @classmethod
    def create_field(cls, 
        table: 'Table', 
        name: str, 
        data_type: 'FieldType', 
        data: Dict = {},
    ) -> 'TableField':
        """
        Create a new field

        Args:
            table (Table): field's table
            name (str): name of the field
            data_type (FieldType): type of the field
            data (Dict, optional): the data of the field

        Returns:
            Field: the created field object

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table
            >>> from geobox.field import Field
            >>> client = GeoboxClient()
            >>> table = client.get_table(uuid="12345678-1234-5678-1234-567812345678")
            >>> field = Field.create_field(client, table=table, name='test', data_type=FieldType.Integer)
        """
        data.update({
            "name": name,
            "datatype": data_type.value
        })
        endpoint = urljoin(table.endpoint, 'fields/')
        return super()._create(table.api, endpoint, data, factory_func=lambda api, item: TableField(table, data_type, item['id'], item))


    def delete(self) -> None:
        """
        Delete the field.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.field import TableField
            >>> client = GeoboxClient()
            >>> table = client.get_table(uuid="12345678-1234-5678-1234-567812345678")
            >>> field = table.get_field(name='test')
            >>> field.delete()
        """
        super()._delete(self.endpoint)
        self.field_id = None


    def update(self, **kwargs) -> Dict:
        """
        Update the field.

        Keyword Args:
            name (str): The name of the field.
            display_name (str): The display name of the field.
            description (str): The description of the field.
            domain (Dict): the domain of the field
            hyperlink (bool): the hyperlink field.

        Returns:
            Dict: The updated data.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.field import TableField
            >>> client = GeoboxClient()
            >>> table = client.get_table(uuid="12345678-1234-5678-1234-567812345678")
            >>> field = table.get_field(name='test')
            >>> field.update(name="my_field", display_name="My Field", description="My Field Description")
        """       
        data = {
            "name": kwargs.get('name'),
            "display_name": kwargs.get('display_name'),
            "description": kwargs.get('description'),
            "domain": kwargs.get('domain'),
            "hyperlink": kwargs.get('hyperlink')
        }
        return super()._update(self.endpoint, data)


    def update_domain(self, 
        range_domain: Dict = None, 
        list_domain: Dict = None,
    ) -> Dict:
        """
        Update field domian values

        Args:
            range_domain (Dict): a dictionary with min and max keys.
            list_domain (Dict): a dictionary containing the domain codes and values.

        Returns:
            Dict: the updated field domain

        Example:
            >>> from geobox import GeoboxClient
            >>> client = GeoboxClient()
            >>> field = client.get_table(uuid="12345678-1234-5678-1234-567812345678").get_fields()[0]
            >>> range_d = {'min': 1, 'max': 10}
            >>> field.update_domain(range_domain = range_d)
            or 
            >>> list_d = {'1': 'value1', '2': 'value2'}
            >>> field.update_domain(list_domain=list_d)
        """
        if not self.domain:
            self.domain = {'min': None, 'max': None, 'items': {}}

        if range_domain:
            self.domain['min'] = range_domain['min']
            self.domain['max'] = range_domain['max']

        if list_domain:
            self.domain['items'] = {**self.domain['items'], **list_domain}

        self.update(domain=self.domain)
        return self.domain


    def to_async(self, async_client: 'AsyncGeoboxClient') -> 'AsyncTableField':
        """
        Switch to async version of the field instance to have access to the async methods

        Args:
            async_client (AsyncGeoboxClient): The async version of the GeoboxClient instance for making requests.

        Returns:
            AsyncField: the async instance of the field.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> client = GeoboxClient()
            >>> table = client.get_table(uuid="12345678-1234-5678-1234-567812345678")
            >>> field = table.get_field(name='test')
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     async_field = field.to_async(async_client)
        """
        from .aio.table import AsyncTableField

        async_table = self.table.to_async(async_client=async_client)
        return AsyncTableField(table=async_table, data_type=self.data_type, field_id=self.field_id, data=self.data)



class Table(Base):

    BASE_ENDPOINT = 'tables/'

    def __init__(self, 
        api: 'GeoboxClient', 
        uuid: str,
        data: Optional[Dict] = {}):
        """
        Initialize a table instance.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            uuid (str): The unique identifier for the table.
            data (Dict): The response data of the table.
        """
        super().__init__(api, uuid=uuid, data=data)


    @classmethod
    def get_tables(cls, api: 'GeoboxClient', **kwargs) -> Union[List['Table'], int]:
        """
        Get list of tables with optional filtering and pagination.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.

        Keyword Args:
            include_settings (bool): Whether to include table settings. default: False
            temporary (bool): Whether to return temporary tables. default: False
            q (str): query filter based on OGC CQL standard. e.g. "field1 LIKE '%GIS%' AND created_at > '2021-01-01'"
            search (str): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value. NOTE: if q param is defined this param will be ignored
            search_fields (str): comma separated list of fields for searching
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            return_count (bool): Whether to return total count. default: False.
            skip (int): Number of items to skip. default: 0
            limit (int): Number of items to return. default: 10
            user_id (int): Specific user. privileges required
            shared (bool): Whether to return shared tables. default: False

        Returns:
            List[Table] | int: A list of table instances or the total number of tables.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table
            >>> client = GeoboxClient()
            >>> tables = client.get_tables(q="name LIKE '%My table%'")
            or
            >>> tables = Table.get_tables(client, q="name LIKE '%My table%'")
        """
        params = {
           'f': 'json',
           'include_settings': kwargs.get('include_settings', False),
           'temporary': kwargs.get('temporary'),
           'q': kwargs.get('q'),
           'search': kwargs.get('search'),
           'search_fields': kwargs.get('search_fields'),
           'order_by': kwargs.get('order_by'),
           'return_count': kwargs.get('return_count', False),
           'skip': kwargs.get('skip', 0),
           'limit': kwargs.get('limit', 10),
           'user_id': kwargs.get('user_id'),
           'shared': kwargs.get('shared', False)
        }
        return super()._get_list(api, cls.BASE_ENDPOINT, params, factory_func=lambda api, item: Table(api, item['uuid'], item))
    

    @classmethod
    def create_table(cls,
        api: 'GeoboxClient', 
        name: str, 
        display_name: Optional[str] = None, 
        description: Optional[str] = None, 
        temporary: bool = False,
        fields: Optional[List[Dict]] = None,
    ) -> 'Table':
        """
        Create a new table.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            name (str): The name of the Table.
            display_name (str, optional): The display name of the table.
            description (str, optional): The description of the table.
            temporary (bool, optional): Whether to create a temporary tables. default: False
            fields (List[Dict], optional): raw table fields. you can use create_field method for simpler and safer field addition. required dictionary keys: name, datatype

        Returns:
            Table: The newly created table instance.

        Raises:
            ValidationError: If the table data is invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table
            >>> client = GeoboxClient()
            >>> table = client.create_table(name="my_table")
            or
            >>> table = Table.create_table(client, name="my_table")            
        """
        data = {
            "name": name,
            "display_name": display_name,
            "description": description,
            "temporary": temporary,
            "fields": fields,
        }
        return super()._create(api, cls.BASE_ENDPOINT, data, factory_func=lambda api, item: Table(api, item['uuid'], item))

    
    @classmethod
    def get_table(cls, api: 'GeoboxClient', uuid: str, user_id: int = None) -> 'Table':
        """
        Get a table by UUID.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            uuid (str): The UUID of the table to get.
            user_id (int): Specific user. privileges required.

        Returns:
            Table: The Table object.

        Raises:
            NotFoundError: If the table with the specified UUID is not found.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table
            >>> client = GeoboxClient()
            >>> table = client.get_table(uuid="12345678-1234-5678-1234-567812345678")
            or
            >>> table = Table.get_table(client, uuid="12345678-1234-5678-1234-567812345678")
        """
        params = {
            'f': 'json',
            'user_id': user_id,
        }
        return super()._get_detail(api, cls.BASE_ENDPOINT, uuid, params, factory_func=lambda api, item: Table(api, item['uuid'], item))


    @classmethod
    def get_table_by_name(cls, api: 'GeoboxClient', name: str, user_id: int = None) -> Union['Table', None]:
        """
        Get a table by name

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            name (str): the name of the table to get
            user_id (int, optional): specific user. privileges required.

        Returns:
            Table | None: returns the table if a table matches the given name, else None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table
            >>> client = GeoboxClient()
            >>> table = client.get_table_by_name(name='test')
            or
            >>> table = Table.get_table_by_name(client, name='test')
        """
        tables = cls.get_tables(api, q=f"name = '{name}'", user_id=user_id)
        if tables and tables[0].name == name:
            return tables[0]
        else:
            return None


    def update(self, **kwargs) -> Dict:
        """
        Update the table.

        Keyword Args:
            name (str): The name of the table.
            display_name (str): The display name of the table.
            description (str): The description of the table.

        Returns:
            Dict: The updated table data.

        Raises:
            ValidationError: If the table data is invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table
            >>> client = GeoboxClient()
            >>> table = Table.get_table(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> table.update_table(display_name="New Display Name")
        """
        data = {
            "name": kwargs.get('name'),
            "display_name": kwargs.get('display_name'),   
            "description": kwargs.get('description'),
        }
        return super()._update(self.endpoint, data)
    

    def delete(self) -> None:
        """
        Delete the Table.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table
            >>> client = GeoboxClient()
            >>> table = Table.get_table(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> table.delete()
        """
        super()._delete(self.endpoint)


    @property
    def settings(self) -> Dict:
        """
        Get the table's settings.
        
        Returns:
            Dict: The table settings.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table
            >>> client = GeoboxClient()
            >>> table = Table.get_table(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>> setting = table.setting
        """
        return super()._get_settings(endpoint=self.endpoint)
    

    def update_settings(self, settings: Dict) -> Dict:
        """
        Update the settings

        settings (Dict): settings dictionary

        Returns:
            Dict: updated settings

        Example:
            >>> from geobox import GeoboxClient
            >>> client = GeoboxClient()
            >>> table1 = client.get_table(uuid="12345678-1234-5678-1234-567812345678")
            >>> table2 = client.get_table(uuid="12345678-1234-5678-1234-567812345678")
            >>> table1.update_settings(table2.settings)
        """
        return super()._set_settings(self.endpoint, settings)


    def get_fields(self) -> List['TableField']:
        """
        Get all fields of the table.
        
        Returns:
            List[TableField]: A list of Field instances representing the table's fields.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table
            >>> client = GeoboxClient()
            >>> table = Table.get_table(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>> fields = table.get_fields()
        """
        endpoint = urljoin(self.endpoint, 'fields/')
        return super()._get_list(api=self.api, 
            endpoint=endpoint,
            factory_func=lambda api, item: TableField(table=self, data_type=FieldType(item['datatype']), field_id=item['id'], data=item))


    def get_field(self, field_id: int) -> 'TableField':
        """
        Get a specific field by ID.
        
        Args:
            field_id (int, optional): The ID of the field to retrieve.

        Returns:
            TableField: The requested field instance.
            
        Raises:
            NotFoundError: If the field with the specified ID is not found. 

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table
            >>> client = GeoboxClient()
            >>> table = Table.get_table(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>> field = table.get_field(field_id=1)
        """        
        field = next((f for f in self.get_fields() if f.id == field_id), None)
        if not field:
            raise NotFoundError(f'Field with ID {field_id} not found in table {self.name}')
            
        return field


    def get_field_by_name(self, name: str) -> 'TableField':
        """
        Get a specific field by name.
        
        Args:
            name (str): The name of the field to retrieve.

        Returns:
            TableField: The requested field instance.
            
        Raises:
            NotFoundError: If the field with the specified name is not found. 

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table
            >>> client = GeoboxClient()
            >>> table = Table.get_table(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>> field = table.get_field_by_name(name='test')
        """        
        field = next((f for f in self.get_fields() if f.name == name), None)
        if not field:
            raise NotFoundError(f"Field with name '{name}' not found in table {self.name}")
            
        return field


    def add_field(self, name: str, data_type: 'FieldType', data: Dict = {}) -> 'TableField':
        """
        Add a new field to the table.
        
        Args:
            name (str): The name of the new field.
            data_type (FieldType): The data type of the new field.
            data (Dict, optional): Additional field properties (display_name, description, etc.).
            
        Returns:
            Field: The newly created field instance.
            
        Raises:
            ValidationError: If the field data is invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table
            >>> client = GeoboxClient()
            >>> table = Table.get_table(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>> field = table.add_field(name="new_field", data_type=FieldType.String)
        """
        return TableField.create_field(table=self, name=name, data_type=data_type, data=data)


    def calculate_field(self, 
        target_field: str, 
        expression: str, 
        q: Optional[str] = None, 
        search: Optional[str] = None, 
        search_fields: Optional[str] = None, 
        row_ids: Optional[str] = None, 
        run_async: bool = True, 
        user_id: Optional[int] = None,
    ) -> Union['Task', Dict]:
        """
        Calculate values for a field based on an expression.
        
        Args:
            target_field (str): The field to calculate values for.
            expression (str): The expression to use for calculation.
            q (str, optional): Query to filter features. default: None.
            search (str, optional): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value
            search_fields (str, optional): comma separated list of fields for searching
            row_ids (str, optional): List of specific row IDs to include. default: None
            run_async (bool, optional): Whether to run the calculation asynchronously. default: True.
            user_id (int, optional): Specific user. privileges required.
            
        Returns:
            Task | Dict: The task instance of the calculation operation or the api response if the run_async=False.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table
            >>> client = GeoboxClient()
            >>> table = Table.get_table(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>> task = table.calculate_field(target_field="target_field", 
            ...     expression="expression", 
            ...     q="name like 'my_layer'", 
            ...     row_ids=[1, 2, 3], 
            ...     run_async=True)
        """
        data = clean_data({
            "target_field": target_field,
            "expression": expression,
            "q": q,
            "search": search,
            "search_fields": search_fields,
            "row_ids": row_ids,
            "run_async": run_async,
            "user_id": user_id
        })
        
        endpoint = urljoin(self.endpoint, 'calculateField/')
        response = self.api.post(endpoint, data, is_json=False)
        if run_async:
            task = Task.get_task(self.api, response.get('task_id'))
            return task

        return response


    def get_rows(self, **kwargs) -> List['TableRow']:
        """
        Query rows of a table

        Keyword Args:
            q (str): Advanced filtering expression, e.g., 'status = "active" and age > 20'
            search (str): Search term for keyword-based searching among fields/columns
            search_fields (str): Comma separated column names to search in
            row_ids (str): Comma separated list of row ids to filter for
            fields (str): Comma separated column names to include in results, or [ALL]
            exclude (str): Comma separated column names to exclude from result
            order_by (str): Comma separated list for ordering, e.g., 'name A, id D'
            skip (int): Number of records to skip for pagination. default: 0
            limit (int): Maximum number of records to return. default: 100
            return_count (bool): If true, returns only the count of matching rows
            user_id (int): Specific user. privileges required

        Returns:
            List[TableRow]: list of table rows objects

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table
            >>> client = GeoboxClient()
            >>> table = client.get_table(uuid="12345678-1234-5678-1234-567812345678")
            or
            >>> table = Table.get_table(client, uuid="12345678-1234-5678-1234-567812345678")

            >>> rows = table.get_rows()
        """
        params = {
            'f': 'json',
            'q': kwargs.get('q'),
            'search': kwargs.get('search'),
            'search_fields': kwargs.get('search_fields'),
            'row_ids': kwargs.get('row_ids'),
            'fields': kwargs.get('fields'),
            'exclude': kwargs.get('exclude'),
            'order_by': kwargs.get('order_by'),
            'skip': kwargs.get('skip', 0),
            'limit': kwargs.get('limit', 100),
            'return_count': kwargs.get('return_count', False),
            'user_id': kwargs.get('user_id'),
        }

        endpoint = f'{self.endpoint}rows/'

        return super()._get_list(api=self.api,
            endpoint=endpoint,
            params=params,
            factory_func=lambda api, item: TableRow(self, item))


    def get_row(self, 
        row_id: int, 
        user_id: Optional[int] = None,
    ) -> 'TableRow':
        """
        Get a row by its id

        Args:
            row_id (int): the row id
            user_id (int, optional): specific user. privileges required.

        Returns:
            TanbleRow: the table row instance

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table
            >>> client = GeoboxClient()
            >>> table = client.get_table(uuid="12345678-1234-5678-1234-567812345678")
            or
            >>> table = Table.get_table(client, uuid="12345678-1234-5678-1234-567812345678")

            >>> row = table.get_row(row_id=1)
        """
        return TableRow.get_row(self, row_id, user_id)


    def create_row(self, **kwargs) -> 'TableRow':
        """
        Create a new row in the table.

        Each keyword argument represents a field value for the row, where:
            - The keyword is the field name
            - The value is the field value

        Keyword Args: 
            **kwargs: Arbitrary field values matching the table schema.

        Returns:
            TableRow: created table row instance

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table
            >>> client = GeoboxClient()
            >>> table = client.get_table(uuid="12345678-1234-5678-1234-567812345678")
            or
            >>> table = Table.get_table(client, uuid="12345678-1234-5678-1234-567812345678")

            >>> row = table.create_row(
            ...     field1=value1
            ... )
        """
        return TableRow.create_row(self, **kwargs)


    def import_rows(self, 
        file: 'File',
        *,
        file_encoding: str = "utf-8",
        input_dataset: Optional[str] = None,
        delimiter: str = ',',
        has_header: bool = True,
        report_errors: bool = False,
        bulk_insert: bool = True,
    ) -> 'Task':
        """
        Import rows from a CSV file into a table
        
        Args:
            file (File): file object to import.
            file_encoding (str, optional): Character encoding of the input file. default: utf-8
            input_dataset (str, optional): Name of the dataset in the input file.
            delimiter (str, optional): the delimiter of the dataset. default: ,
            has_header (bool, optional): Whether the file has header or not. default: True
            report_errors (bool, optional): Whether to report import errors. default: False
            bulk_insert (bool, optional): 
            
        Returns:
            Task: The task instance of the import operation.
            
        Raises:
            ValidationError: If the import parameters are invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table
            >>> client = GeoboxClient()
            >>> table = Table.get_table(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>> file = client.get_file(uuid="12345678-1234-5678-1234-567812345678")
            >>> task = table.import_rows(
            ...     file=file,
            ... )
        """
        data = clean_data({
            "file_uuid": file.uuid,
            "file_encoding": file_encoding,
            "input_dataset": file.name if not input_dataset else input_dataset,
            "delimiter": delimiter,
            "has_header": has_header,
            "report_errors": report_errors,
            "bulk_insert": bulk_insert,
        })

        endpoint = urljoin(self.endpoint, 'import-csv/')
        response = self.api.post(endpoint, data, is_json=False)
        task = Task.get_task(self.api, response.get('task_id'))
        return task


    def export_rows(self,
        out_filename: str,
        *,
        out_format: 'TableExportFormat' = TableExportFormat.CSV, 
        q: Optional[str] = None,
        search: Optional[str] = None,
        search_fields: Optional[str] = None,
        row_ids: Optional[str] = None,
        fields: Optional[str] = None,
        exclude: Optional[str] = None,
        order_by: Optional[str] = None,
        zipped: bool = False,
        run_async: bool = True,
    ) -> Union['Task', str]:
        """
        Export rows of a table to a file
        
        Args:
            out_filename (str): Name of the output file without the format (.csv) 
            out_format (TableExportFormat, optional): Format of the output file
            q (str, optional): query filter based on OGC CQL standard. e.g. "field1 LIKE '%GIS%' AND created_at > '2021-01-01'"
            search (str, optional): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value
            search_fields (str, optional): comma separated list of fields for searching
            row_ids (str, optional): List of specific row IDs to include
            fields (str, optional): List of specific field names to include
            exclude (str, optional): List of specific field names to exclude
            order_by (str, optional): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            zipped (str, optional): Whether to compress the output file
            run_async (bool, optional): Whether to run the export asynchronously. default: True
            
        Returns:
            Task | Dict: The task instance of the export operation (run_async=True) or the export result (run_async=False)
            
        Raises:
            ValidationError: If the export parameters are invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table
            >>> client = GeoboxClient()
            >>> table = Table.get_table(api=client, uuid="12345678-1234-5678-1234-567812345678")
            >>> file = client.get_file(uuid="12345678-1234-5678-1234-567812345678")
            >>> task = table.export_rows(
            ...     file=file,
            ... )
        """
        data = clean_data({
            "out_filename": out_filename,
            "out_format": out_format.value if out_format else None,
            "q": q,
            "search": search,
            "search_fields": search_fields,
            "row_ids": row_ids,
            "fields": fields,
            "exclude": exclude,
            "order_by": order_by,
            "zipped": zipped,
            "run_async": run_async,
        })

        endpoint = urljoin(self.endpoint, 'export/')
        response = self.api.post(endpoint, data, is_json=False)
        if run_async:
            task = Task.get_task(self.api, response.get('task_id'))
            return task

        return response


    def share(self, users: List['User']) -> None:
        """
        Shares the table with specified users.

        Args:
            users (List[User]): The list of user objects to share the table with.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient             
            >>> from geobox.table import Table
            >>> client = GeoboxClient()
            >>> table = Table.get_table(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> users = client.search_users(search='John')
            >>> table.share(users=users)
        """
        super()._share(self.endpoint, users)
    

    def unshare(self, users: List['User']) -> None:
        """
        Unshares the table with specified users.

        Args:
            users (List[User]): The list of user objects to unshare the table with.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient             
            >>> from geobox.table import Table
            >>> client = GeoboxClient()
            >>> table = Table.get_table(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> users = client.search_users(search='John')
            >>> table.unshare(users=users)
        """
        super()._unshare(self.endpoint, users)


    def get_shared_users(self, search: str = None, skip: int = 0, limit: int = 10) -> List['User']:
        """
        Retrieves the list of users the table is shared with.

        Args:
            search (str, optional): The search query.
            skip (int, optional): The number of users to skip.
            limit (int, optional): The maximum number of users to retrieve.

        Returns:
            List[User]: The list of shared users.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.table import Table
            >>> client = GeoboxClient()
            >>> table = Table.get_table(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> table.get_shared_users(search='John', skip=0, limit=10)
        """
        params = {
            'search': search,
            'skip': skip,
            'limit': limit
        }
        return super()._get_shared_users(self.endpoint, params)


    def to_async(self, async_client: 'AsyncGeoboxClient') -> 'AsyncTable':
        """
        Switch to async version of the table instance to have access to the async methods

        Args:
            async_client (AsyncGeoboxClient): The async version of the GeoboxClient instance for making requests.

        Returns:
            AsyncTable: the async instance of the table.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> client = GeoboxClient()
            >>> table = Table.get_table(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     async_table = table.to_async(async_client)
        """
        from .aio.table import AsyncTable

        return AsyncTable(api=async_client, uuid=self.uuid, data=self.data)


import pytest

from geobox.enums import GroupByAggFunction, NetworkTraceDirection, SpatialAggFunction, SpatialPredicate
from geobox.exception import NotFoundError
from geobox.task import Task
from geobox.vector_tool import VectorTool
from geobox.query import Query

def test_init(api):
    vector_tool = VectorTool(api)
    assert type(vector_tool) == VectorTool
    assert vector_tool.api == api


def test_repr(api):
    vector_tool = VectorTool(api)
    assert repr(vector_tool) == 'VectorTool()'


def test_wrong_query_name(api):
    vector_tool = VectorTool(api)
    api.get_system_queries.return_value = []
    with pytest.raises(NotFoundError):
        vector_tool.area(
            vector_uuid='1234'
        )


def test_area_execute(api):
    vector_tool = VectorTool(api)
    # execute
    output = {'result': 'test'}
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': '$area',
            'sql': 'test',
            "params": [
                {
                "name": "layer",
                "type": "Layer",
                "default_value": None,
                "domain": None
                },
            ]
        })]
    api.post.return_value = output
    result = vector_tool.area(
        vector_uuid='1234'
    )

    assert result == output


def test_area_save_as_layer(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': '$area',
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.area(
        vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_as_geojson(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$as_geojson",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.as_geojson(
        vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_as_wkt(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$as_wkt",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.as_wkt(
        vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_buffer(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$buffer",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.buffer(
        vector_uuid='1234',
        distance=20,
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_centroid(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$centroid",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.centroid(
        vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_clip(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$clip",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.clip(
        vector_uuid='1234',
        clip_vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_concave_hull(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$concave_hull",
            'sql': 'test'
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.concave_hull(
        vector_uuid='1234',
        tolerance=10,
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_convex_hull(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$convex_hull",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.convex_hull(
        vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_feature_count_execute(api):
    vector_tool = VectorTool(api)
    # execute
    output = {'result': 'test'}
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': '$count',
            'sql': 'test',
            "params": [
                {
                "name": "layer",
                "type": "Layer",
                "default_value": None,
                "domain": None
                },
            ]
        })]
    api.post.return_value = output
    result = vector_tool.feature_count(
        vector_uuid='1234'
    )

    assert result == output


def test_count_point_in_polygons(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$count_point_in_polygons",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.count_point_in_polygons(
        polygon_vector_uuid='1234',
        point_vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_delaunay_triangulation(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$delaunay_triangulation",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.delaunay_triangulation(
        vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_disjoint(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$disjoint",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.disjoint(
        vector_uuid='1234',
        filter_vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_dissolve(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$dissolve",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.dissolve(
        vector_uuid='1234',
        dissolve_field_name='field1',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_distance_to_nearest(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$distance_to_nearest",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.distance_to_nearest(
        vector_uuid='1234',
        nearest_vector_uuid='1234',
        search_radius=10,
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_distinct(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$distinct",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.distinct(
        vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_dump(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$dump",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.dump(
        vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_erase(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$erase",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.erase(
        vector_uuid='1234',
        erase_vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_feature_bbox(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$feature_bbox",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.feature_bbox(
        vector_uuid='1234',
        srid=1234,
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_feature_extent(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$feature_extent",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.feature_extent(
        vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_find_and_replace(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$find_replace",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.find_and_replace(
        vector_uuid='1234',
        find="find",
        replace="replace",
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_from_geojson(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$from_geojson",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.from_geojson(
        vector_uuid='1234',
        json_field_name='field',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_from_wkt(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$from_wkt",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.from_wkt(
        vector_uuid='1234',
        wkt_column_name='field',
        srid=1234,
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_generate_points(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$generate_points",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.generate_points(
        vector_uuid='1234',
        number_of_points=10,
        seed=1234,
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_group_by(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$group_by",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.group_by(
        vector_uuid='1234',
        group_column_name='field1',
        agg_column_name='field2',
        agg_function=GroupByAggFunction.COUNT,
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_hexagon_grid(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$hexagon_grid",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.hexagon_grid(
        vector_uuid='1234',
        cell_size=10,
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_intersection(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$intersection",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.intersection(
        vector_uuid='1234',
        intersect_vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_join(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$join",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.join(
        vector1_uuid='1234',
        vector2_uuid='1234',
        vector1_join_column='field',
        vector2_join_column='field',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_lat_lon_to_point(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$lat_lon_to_point",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.lat_lon_to_point(
        latitude=10,
        longitude=10,
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_layer_bbox(api):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$layer_bbox",
            'sql': 'test',
              "params": [
                {
                "name": "layer",
                "type": "Layer",
                "default_value": None,
                "domain": None
                },
                {
                "name": "srid",
                "type": "Integer",
                "default_value": 4326,
                "domain": {
                    "min": 1,
                    "max": 65536,
                    "items": None
                }
                }
            ],
        })]
    api.post.return_value = {'data': 'data'}
    result = vector_tool.layer_bbox(
        vector_uuid='1234',
        srid=1234
    )

    assert type(result) == dict


def test_layer_extent(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$layer_extent",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.layer_extent(
        vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_length(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$length",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.length(
        vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_linear_referencing(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$linear_referencing",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.linear_referencing(
        vector_uuid='1234',
        feature_id=1,
        dist=10,
        tolerance=10,
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_line_to_polygon(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$line_to_polygon",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.line_to_polygon(
        vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_make_envelop(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$make_envelop",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.make_envelop(
        xmin=1,
        ymin=2,
        xmax=3,
        ymax=4,
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_merge(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$merge",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.merge(
        vector1_uuid='1234',
        vector2_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_network_trace(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$network_trace",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.network_trace(
        vector_uuid='1234',
        from_id=1,
        direction=NetworkTraceDirection.UP,
        tolerance=10,
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_number_of_geoms(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$number_of_geoms",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.number_of_geoms(
        vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_number_of_points(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$number_of_points",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.number_of_points(
        vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_point_stats_in_polygon(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$point_stats_in_polygon",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.point_stats_in_polygon(
        polygon_vector_uuid='1234', 
        point_vector_uuid='1234', 
        stats_column='field', 
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_remove_holes(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$remove_holes",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.remove_holes(
        vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_reverse_line(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$reverse_line",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.reverse_line(
        vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_simplify(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$simplify",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.simplify(
        vector_uuid='1234',
        tolerance=10,
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_snap_to_grid(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$snap_to_grid",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.snap_to_grid(
        vector_uuid='1234',
        grid_size=10,
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_spatial_aggregation(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$spatial_aggregation",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.spatial_aggregation(
        vector_uuid='1234',
        agg_function=SpatialAggFunction.COLLECT,
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_spatial_filter(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$spatial_filter",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.spatial_filter(
        vector_uuid='1234',
        spatial_predicate=SpatialPredicate.CONTAIN,
        filter_vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_spatial_group_by(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$spatial_group_by",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.spatial_group_by(
        vector_uuid='1234',
        group_column_name='field',
        agg_function=SpatialAggFunction.COLLECT,
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_square_grid(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$square_grid",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.square_grid(
        vector_uuid='1234',
        cell_size=10,
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_voronoi_polygons(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$voronoi_polygons",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.voronoi_polygons(
        vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_within_distance(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$within_distance",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.within_distance(
        vector_uuid='1234',
        filter_vector_uuid='1234',
        dist=10,
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


def test_xy_coordinate(api, mock_success_task_data):
    vector_tool = VectorTool(api)
    # execute
    api.get_system_queries.return_value = [Query(api=api, uuid='1234', 
        data={
            'name': "$xy_coordinate",
            'sql': 'test',
        })]
    api.post.return_value = {'task_id': mock_success_task_data['id']}
    api.get.return_value = mock_success_task_data
    result = vector_tool.xy_coordinate(
        vector_uuid='1234',
        output_layer_name='test'
    )

    assert type(result) == Task
    assert result.uuid == mock_success_task_data['uuid']


from unittest.mock import patch

from geobox.mosaic import Mosaic
from geobox.raster import Raster
from geobox.task import Task
from geobox.user import User
from geobox.utils import clean_data


def test_init(api, mock_mosaic_data):
    """Test the initialization of a Mosaic object."""
    mosaic = Mosaic(api, mock_mosaic_data['uuid'], mock_mosaic_data)
    assert mosaic.uuid == mock_mosaic_data['uuid']
    assert mosaic.name == mock_mosaic_data['name']
    assert mosaic.data == mock_mosaic_data


def test_get_mosaics(api, mock_mosaic_data):
    """Test the get_mosaics method."""
    mock_response = [mock_mosaic_data, mock_mosaic_data]
    api.get.return_value = mock_response
    
    mosaics = Mosaic.get_mosaics(api)
    
    api.get.assert_called_once_with('mosaics/?f=json&return_count=False&skip=0&limit=100&shared=False')
    assert len(mosaics) == 2
    assert all(isinstance(m, Mosaic) for m in mosaics)
    assert mosaics[0].name == mock_mosaic_data['name']
    assert mosaics[1].data == mock_mosaic_data
    assert mosaics[0].uuid == mock_mosaic_data['uuid']


def test_get_mosaic(api, mock_mosaic_data):
    """Test the get_mosaic method."""
    api.get.return_value = mock_mosaic_data
    
    mosaic = Mosaic.get_mosaic(api, mock_mosaic_data['uuid'])
    
    api.get.assert_called_once_with(f'mosaics/{mock_mosaic_data["uuid"]}/?f=json')
    assert isinstance(mosaic, Mosaic)
    assert mosaic.uuid == mock_mosaic_data['uuid']
    assert mosaic.name == mock_mosaic_data['name']


def test_get_mosaic_by_name(api, mock_mosaic_data):
    """Test the get_mosaic method."""
    api.get.return_value = [mock_mosaic_data]
    
    mosaic = Mosaic.get_mosaic_by_name(api, name=mock_mosaic_data['name'])
    
    api.get.assert_called_once_with(f'{Mosaic.BASE_ENDPOINT}?f=json&q=name+%3D+%27string%27&return_count=False&skip=0&limit=100&shared=False')
    assert isinstance(mosaic, Mosaic)
    assert mosaic.uuid == mock_mosaic_data['uuid']
    assert mosaic.name == mock_mosaic_data['name']
    # not found
    mosaic = Mosaic.get_mosaic_by_name(api, name='not_found')
    assert mosaic == None


def test_create_mosaic(api, mock_mosaic_data):
    """Test the create_mosaic method."""
    api.post.return_value = mock_mosaic_data
    
    mosaic = Mosaic.create_mosaic(
        api,
        name=mock_mosaic_data['name'],
        display_name=mock_mosaic_data['display_name'],
        description=mock_mosaic_data['description'],
        pixel_selection=mock_mosaic_data['pixel_selection'],
        min_zoom=mock_mosaic_data['min_zoom']
    )
    
    api.post.assert_called_once_with('mosaics/', {
        'name': mock_mosaic_data['name'],
        'display_name': mock_mosaic_data['display_name'],
        'description': mock_mosaic_data['description'],
        'pixel_selection': mock_mosaic_data['pixel_selection'],
        'min_zoom': mock_mosaic_data['min_zoom']
    })
    assert isinstance(mosaic, Mosaic)
    assert mosaic.uuid == mock_mosaic_data['uuid']
    assert mosaic.name == mock_mosaic_data['name']


def test_get_mosaics_by_ids(api, mock_mosaic_data):
    """Test the get_mosaics_by_ids method."""
    mock_response = [{**mock_mosaic_data, **{'id': 1}}, {**mock_mosaic_data, **{'id': 2}}]
    api.get.return_value = mock_response
    
    mosaics = Mosaic.get_mosaics_by_ids(api, [1, 2])
    
    api.get.assert_called_once_with('mosaics/get-mosaics/?ids=%5B1%2C+2%5D')
    assert len(mosaics) == 2
    assert all(isinstance(m, Mosaic) for m in mosaics)
    assert mosaics[0].data == {**mock_mosaic_data, **{'id': 1}}
    assert mosaics[1].data == {**mock_mosaic_data, **{'id': 2}}


def test_update(api, mock_mosaic_data):
    """Test the update_mosaic method."""
    mosaic = Mosaic(api, mock_mosaic_data['uuid'], mock_mosaic_data)
    updated_data = {**mock_mosaic_data, 'name': 'updated_name'}
    api.put.return_value = updated_data
    
    mosaic.update(name='updated_name')
    
    api.put.assert_called_once_with(f'mosaics/{mock_mosaic_data["uuid"]}/', {'name': 'updated_name'})
    assert mosaic.name == 'updated_name'


def test_delete(api, mock_mosaic_data):
    """Test the delete_mosaic method."""
    mosaic = Mosaic(api, mock_mosaic_data['uuid'], mock_mosaic_data)
    endpoint = mosaic.endpoint
    mosaic.delete()
    
    api.delete.assert_called_once_with(endpoint)
    assert mosaic.uuid is None
    assert mosaic.endpoint is None


def test_thumbnail(api, mock_mosaic_data):
    """Test the get_thumbnail method."""
    mosaic = Mosaic(api, mock_mosaic_data['uuid'], mock_mosaic_data)
    
    thumbnail = mosaic.thumbnail
    
    assert thumbnail == f"{api.base_url}{mosaic.endpoint}thumbnail"


def test_get_point(api, mock_mosaic_data):
    """Test the get_point method."""
    mosaic = Mosaic(api, mock_mosaic_data['uuid'], mock_mosaic_data)
    api.get.return_value = {'point': 'value'}
    
    point = mosaic.get_point(1, 2)

    assert point == {'point': 'value'}
    api.get.assert_called_once_with(f'{mosaic.endpoint}point?lat=1&lng=2')
    

def test_get_render_png_url(api, mock_mosaic_data):
    """Test the get_tile_render_url method."""
    mosaic = Mosaic(api, mock_mosaic_data['uuid'], mock_mosaic_data)
    
    tile_render_url = mosaic.get_render_png_url(x=1, y=2, z=3)

    assert tile_render_url == f'{api.base_url}{mosaic.endpoint}render/3/1/2.png'


def test_get_tile_png_url(api, mock_mosaic_data):
    """Test the get_tile_png_url method."""
    mosaic = Mosaic(api, mock_mosaic_data['uuid'], mock_mosaic_data)
    
    tile_png_url = mosaic.get_tile_png_url(x=1, y=2, z=3)

    assert tile_png_url == f'{api.base_url}{mosaic.endpoint}tiles/3/1/2.png'
    

def test_get_tile_json(api, mock_mosaic_data):
    """Test the get_tile_json method."""
    mosaic = Mosaic(api, mock_mosaic_data['uuid'], mock_mosaic_data)
    api.get.return_value = {'value': 'test'}
    result = mosaic.get_tile_json()    
    api.get.assert_called_once_with(f'{mosaic.endpoint}tilejson.json')
    assert result == {'value': 'test'}


def test_wmts(api, mock_mosaic_data):
    """Test the get_wmts method."""
    mosaic = Mosaic(api, mock_mosaic_data['uuid'], mock_mosaic_data)
    wmts_url = mosaic.wmts(scale=1)
    assert wmts_url == f'{api.base_url}{mosaic.endpoint}wmts/?scale=1'


def test_settings(api, mock_mosaic_data):
    """Test the get_settings method."""
    mosaic = Mosaic(api, mock_mosaic_data['uuid'], mock_mosaic_data)
    api.get.return_value = {'value': 'test'}
    
    settings = mosaic.settings

    api.get.assert_called_once_with(f'{mosaic.endpoint}settings/?f=json')
    assert settings == {'value': 'test'}


def test_update_settings(api, mock_mosaic_data):
    mosaic1 = Mosaic(api, uuid=mock_mosaic_data['uuid'], data=mock_mosaic_data)
    mosaic2 = Mosaic(api, uuid=mock_mosaic_data['uuid'], data=mock_mosaic_data)
    settings = mosaic2.settings
    api.put.return_value = settings
    mosaic1.update_settings(mosaic2.settings)
    api.put.assert_called_once_with(f'{mosaic1.endpoint}settings/', settings)


def test_set_settings(api, mock_mosaic_data):
    """Test the set_settings method."""
    mosaic = Mosaic(api, mock_mosaic_data['uuid'], mock_mosaic_data)
    api.put.return_value = 'test-settings-url'
    api.get.return_value = {
        "visual_settings": {
            "nodata": None,
            "indexes": [
            1,
            2,
            3
            ],
            "rescale": [
            [
                0,
                9848
            ],
            [
                0,
                9636
            ],
            [
                0,
                9409
            ]
            ],
            "colormap_name": None,
            "colormap": None,
            "color_formula": None,
            "expression": None,
            "exaggeration": 10
        },
        "tile_settings": {
            "min_zoom": 0,
            "max_zoom": 13,
            "use_cache": True,
            "cache_until_zoom": 17
        }
        }
    
    mosaic.set_settings(nodata=0, 
                        indexes=[1], 
                        rescale=[[0, 10000]], 
                        colormap_name='gist_rainbow', 
                        color_formula='Gamma R 0.5', 
                        expression='b1 * 2', 
                        exaggeraion=10, 
                        min_zoom=0, 
                        max_zoom=22, 
                        use_cache=True, 
                        cache_until_zoom=17)
    expected_data = {'visual_settings': 
                        {'nodata': 0, 
                        'indexes': [1], 
                        'rescale': [[0, 10000]], 
                        'colormap_name': 'gist_rainbow', 
                        'colormap': None, 
                        'color_formula': 'Gamma R 0.5', 
                        'expression': 'b1 * 2', 
                        'exaggeration': 10, 'exaggeraion': 10}, 
                    'tile_settings': 
                        {'min_zoom': 0, 
                        'max_zoom': 22, 
                        'use_cache': True, 
                        'cache_until_zoom': 17}}
    api.put.assert_called_once_with(f'{mosaic.endpoint}settings/', expected_data)


def test_get_rasters(api, mock_mosaic_data, mock_raster_data):
    """Test the get_rasters method."""
    mosaic = Mosaic(api, mock_mosaic_data['uuid'], mock_mosaic_data)
    api.get.return_value = [mock_raster_data, mock_raster_data]
    
    rasters = mosaic.get_rasters()
    
    api.get.assert_called_once_with(f'{mosaic.endpoint}rasters')
    assert len(rasters) == 2
    assert all(isinstance(r, Raster) for r in rasters)
    assert rasters[0].name == mock_raster_data['name']
    assert rasters[1].data == mock_raster_data


def test_add_rasters(api, mock_mosaic_data, mock_raster_data):
    """Test the add_raster method."""
    mosaic = Mosaic(api, mock_mosaic_data['uuid'], mock_mosaic_data)
    rasters = [
        Raster(api, mock_raster_data['uuid'], mock_raster_data),
        Raster(api, mock_raster_data['uuid'], mock_raster_data),
        Raster(api, mock_raster_data['uuid'], mock_raster_data),
    ]
    rasters[0].id = 1
    rasters[1].id = 2
    rasters[2].id = 3
    mosaic.add_rasters(rasters=rasters)
    
    api.post.assert_called_once_with(
        f'{mosaic.endpoint}rasters/',
        {'raster_ids': [1, 2, 3]},
        is_json=False
    )


def test_remove_rasters(api, mock_mosaic_data, mock_raster_data):
    """Test the remove_raster method."""
    mosaic = Mosaic(api, mock_mosaic_data['uuid'], mock_mosaic_data)
    rasters = [
        Raster(api, mock_raster_data['uuid'], mock_raster_data),
        Raster(api, mock_raster_data['uuid'], mock_raster_data),
        Raster(api, mock_raster_data['uuid'], mock_raster_data),
    ]
    rasters[0].id = 1
    rasters[1].id = 2
    rasters[2].id = 3
    mosaic.remove_rasters(rasters=rasters)
    
    api.delete.assert_called_once_with(
        f'{mosaic.endpoint}rasters/?raster_ids=%5B1%2C+2%2C+3%5D',
        is_json=False
    )


def test_share(api, mock_mosaic_data, mock_user_data):
    """Test file sharing."""
    with patch.object(api, 'post'):
        mosaic = Mosaic(api, uuid=mock_mosaic_data['uuid'], data=mock_mosaic_data)
        users = [
            User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1']),
            User(api, user_id=mock_user_data['2']['id'], data=mock_user_data['2'])
        ]
        mosaic.share(users=users)
        api.post.assert_called_once_with(
            f'{mosaic.endpoint}share/',
            {'user_ids': [1, 2]}, is_json=False
        )


def test_unshare(api, mock_mosaic_data, mock_user_data):
    """Test file unsharing."""
    with patch.object(api, 'post'):
        mosaic = Mosaic(api, uuid=mock_mosaic_data['uuid'], data=mock_mosaic_data)
        users = [
            User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1']),
            User(api, user_id=mock_user_data['2']['id'], data=mock_user_data['2'])
        ]
        mosaic.unshare(users=users)
        api.post.assert_called_once_with(
            f'{mosaic.endpoint}unshare/',
            {'user_ids': [1, 2]}, is_json=False
        )


def test_get_shared_users(api, mock_mosaic_data, mock_user_data):
    """Test getting shared users."""
    mock_response = [
        mock_user_data['1'],
        mock_user_data['2']
    ]
    
    with patch.object(api, 'get', return_value=mock_response):
        mosaic = Mosaic(api, uuid=mock_mosaic_data['uuid'], data=mock_mosaic_data)
        result = mosaic.get_shared_users(search='user', limit=2)

        api.get.assert_called_once_with(
            f'{mosaic.endpoint}shared-with-users/?search=user&skip=0&limit=2'
        )
        
        assert len(result) == 2
        assert result[0].first_name == 'test 1'
        assert result[0].last_name == 'test 1'
        assert result[1].first_name == 'test 2'
        assert result[1].last_name == 'test 2'


def test_seed_cache(api, mock_mosaic_data, mock_success_task_data):
    """Test the seed_cache method."""
    mosaic = Mosaic(api, mock_mosaic_data['uuid'], mock_mosaic_data)
    api.post.return_value = [{'task_id': mock_success_task_data['id']}]
    api.get_task.return_value = Task(api, mock_success_task_data['uuid'], mock_success_task_data)
    task = mosaic.seed_cache(from_zoom=0, to_zoom=22, workers=1)[0]

    expected_data = clean_data({
        'from_zoom': 0,
        'to_zoom': 22,
        'extent': None,
        'workers': 1
    })
    
    api.post.assert_called_once_with(f'{mosaic.endpoint}cache/seed/', expected_data)
    assert isinstance(task, Task)


def test_clear_cache(api, mock_mosaic_data):
    """Test the clear cache method."""
    mosaic = Mosaic(api, mock_mosaic_data['uuid'], mock_mosaic_data)
    mosaic.clear_cache()
    api.post.assert_called_once_with(f'{mosaic.endpoint}cache/clear/')


def test_cache_size(api, mock_mosaic_data):
    """Test the cache size method."""
    mosaic = Mosaic(api, mock_mosaic_data['uuid'], mock_mosaic_data)
    api.post.return_value = 1024
    size = mosaic.cache_size
    assert size == 1024
    api.post.assert_called_once_with(f'{mosaic.endpoint}cache/size/')


def test_to_async(api, async_api, mock_mosaic_data):
    mosaic = Mosaic(api, mock_mosaic_data['uuid'], mock_mosaic_data)
    async_instance = mosaic.to_async(async_api)
    assert async_instance.api == async_api 


from typing import List, Dict, Optional, TYPE_CHECKING, Union
from urllib.parse import urljoin

from .base import Base

if TYPE_CHECKING:
    from . import GeoboxClient
    from .user import User 
    from .aio import AsyncGeoboxClient
    from .aio.scene import AsyncScene

class Scene(Base):

    BASE_ENDPOINT = 'scenes/'

    def __init__(self, 
                 api: 'GeoboxClient', 
                 uuid: str,
                 data: Optional[Dict] = {}):
        """
        Initialize a Scene instance.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            uuid (str): The unique identifier for the Scene.
            data (Dict): The data of the Scene.
        """
        super().__init__(api, uuid=uuid, data=data)


    @classmethod
    def get_scenes(cls, api: 'GeoboxClient', **kwargs) -> Union[List['Scene'], int]:
        """
        Get list of scenes with optional filtering and pagination.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.

        Keyword Args:
            q (str): query filter based on OGC CQL standard. e.g. "field1 LIKE '%GIS%' AND created_at > '2021-01-01'"
            search (str): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value. NOTE: if q param is defined this param will be ignored.
            search_fields (str): comma separated list of fields for searching.
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            return_count (bool): Whether to return total count. default is False.
            skip (int): Number of items to skip. default is 0.
            limit (int): Number of items to return. default is 10.
            user_id (int): Specific user. privileges required.
            shared (bool): Whether to return shared scenes. default is False.

        Returns:
            List[Scene] | int: A list of scene instances or the total number of scenes.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.scene import Scene
            >>> client = GeoboxClient()
            >>> scenes = Scene.get_scenes(client, q="name LIKE '%My scene%'")
            or
            >>> scenes = client.get_scenes(q="name LIKE '%My scene%'")
        """
        params = {
           'f': 'json',
           'q': kwargs.get('q'),
           'search': kwargs.get('search'),
           'search_fields': kwargs.get('search_fields'),
           'order_by': kwargs.get('order_by'),
           'return_count': kwargs.get('return_count', False),
           'skip': kwargs.get('skip', 0),
           'limit': kwargs.get('limit', 10),
           'user_id': kwargs.get('user_id'),
           'shared': kwargs.get('shared', False)
        }
        return super()._get_list(api, cls.BASE_ENDPOINT, params, factory_func=lambda api, item: Scene(api, item['uuid'], item))
    

    @classmethod
    def create_scene(cls, 
                     api: 'GeoboxClient', 
                     name: str, 
                     display_name: str = None, 
                     description: str = None, 
                     settings: Dict = {}, 
                     thumbnail: str = None, 
                     user_id: int = None) -> 'Scene':
        """
        Create a new scene.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            name (str): The name of the scene.
            display_name (str, optional): The display name of the scene.
            description (str, optional): The description of the scene.
            settings (Dict,optional): The settings of the scene.
            thumbnail (str, optional): The thumbnail of the scene.
            user_id (int, optional): Specific user. privileges required.

        Returns:
            Scene: The newly created scene instance.

        Raises:
            ValidationError: If the scene data is invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.scene import Scene
            >>> client = GeoboxClient()
            >>> scene = Scene.create_scene(client, name="my_scene")
            or
            >>> scene = client.create_scene(name="my_scene")
        """
        data = {
            "name": name,
            "display_name": display_name,
            "description": description,
            "settings": settings,
            "thumbnail": thumbnail,
            "user_id": user_id,
        }
        return super()._create(api, cls.BASE_ENDPOINT, data, factory_func=lambda api, item: Scene(api, item['uuid'], item))

    
    @classmethod
    def get_scene(cls, api: 'GeoboxClient', uuid: str, user_id: int = None) -> 'Scene':
        """
        Get a scene by its UUID.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            uuid (str): The UUID of the scene to get.
            user_id (int, optional): Specific user. privileges required.

        Returns:
            Scene: The scene object.

        Raises:
            NotFoundError: If the scene with the specified UUID is not found.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.scene import Scene
            >>> client = GeoboxClient()
            >>> scene = Scene.get_scene(client, uuid="12345678-1234-5678-1234-567812345678")
            or
            >>> scene = client.get_scene(uuid="12345678-1234-5678-1234-567812345678")
        """
        params = {
            'f': 'json',
            'user_id': user_id,
        }
        return super()._get_detail(api, cls.BASE_ENDPOINT, uuid, params, factory_func=lambda api, item: Scene(api, item['uuid'], item))


    @classmethod
    def get_scene_by_name(cls, api: 'GeoboxClient', name: str, user_id: int = None) -> Union['Scene', None]:
        """
        Get a scene by name

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            name (str): the name of the scene to get
            user_id (int, optional): specific user. privileges required.

        Returns:
            Scene | None: returns the scene if a scene matches the given name, else None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.scene import Scene
            >>> client = GeoboxClient()
            >>> scene = Scene.get_scene_by_name(client, name='test')
            or
            >>> scene = client.get_scene_by_name(name='test')
        """
        scenes = cls.get_scenes(api, q=f"name = '{name}'", user_id=user_id)
        if scenes and scenes[0].name == name:
            return scenes[0]
        else:
            return None


    def update(self, **kwargs) -> Dict:
        """
        Update the scene.

        Keyword Args:
            name (str): The name of the scene.
            display_name (str): The display name of the scene.
            description (str): The description of the scene.
            settings (Dict): The settings of the scene.
            thumbnail (str): The thumbnail of the scene.

        Returns:
            Dict: The updated scene data.

        Raises:
            ApiRequestError: If the API request fails.
            ValidationError: If the scene data is invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.scene import Scene
            >>> client = GeoboxClient()
            >>> scene = Scene.get_scene(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> scene.update(display_name="New Display Name")
        """
        data = {
            "name": kwargs.get('name'),
            "display_name": kwargs.get('display_name'),   
            "description": kwargs.get('description'),
            "settings": kwargs.get('settings'),
            "thumbnail": kwargs.get('thumbnail')
        }
        return super()._update(self.endpoint, data)
    

    def delete(self) -> None:
        """
        Delete the scene.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.scene import Scene
            >>> client = GeoboxClient()
            >>> scene = Scene.get_scene(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> scene.delete()
        """
        super()._delete(self.endpoint)


    @property
    def thumbnail(self) -> str:
        """
        Get the thumbnail URL of the scene.

        Returns:
            str: The thumbnail of the scene.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.scene import Scene
            >>> client = GeoboxClient()
            >>> scene = Scene.get_scene(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> scene.thumbnail
        """
        return super()._thumbnail()
    

    def share(self, users: List['User']) -> None:
        """
        Shares the scene with specified users.

        Args:
            users (List[User]): The list of user objects to share the scene with.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient             
            >>> from geobox.scene import Scene
            >>> client = GeoboxClient()
            >>> scene = Scene.get_scene(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> users = client.search_users(search='John')
            >>> scene.share(users=users)
        """
        super()._share(self.endpoint, users)
    

    def unshare(self, users: List['User']) -> None:
        """
        Unshares the scene with specified users.

        Args:
            users (List[User]): The list of user objects to unshare the scene with.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient             
            >>> from geobox.scene import Scene
            >>> client = GeoboxClient()
            >>> scene = Scene.get_scene(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> users = client.search_users(search='John')
            >>> scene.unshare(users=users)
        """
        super()._unshare(self.endpoint, users)


    def get_shared_users(self, search: str = None, skip: int = 0, limit: int = 10) -> List['User']:
        """
        Retrieves the list of users the scene is shared with.

        Args:
            search (str, optional): The search query.
            skip (int, optional): The number of users to skip.
            limit (int, optional): The maximum number of users to retrieve.

        Returns:
            List[User]: The list of shared users.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.scene import Scene
            >>> client = GeoboxClient()
            >>> scene = Scene.get_scene(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> scene.get_shared_users(search='John', skip=0, limit=10)
        """
        params = {
            'search': search,
            'skip': skip,
            'limit': limit
        }
        return super()._get_shared_users(self.endpoint, params)
    

    def to_async(self, async_client: 'AsyncGeoboxClient') -> 'AsyncScene':
        """
        Switch to async version of the scene instance to have access to the async methods

        Args:
            async_client (AsyncGeoboxClient): The async version of the GeoboxClient instance for making requests.

        Returns:
            AsyncScene: the async instance of the scene.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.scene import Scene
            >>> client = GeoboxClient()
            >>> scene = Scene.get_scene(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     async_scene = scene.to_async(async_client)
        """
        from .aio.scene import AsyncScene

        return AsyncScene(api=async_client, uuid=self.uuid, data=self.data)
